import os

import numpy as np

from pathlib import Path

from scipy.spatial.distance import euclidean

from boss.bo.acq.explore import explore
from boss.bo.hmc import HMCwrapper
from boss.bo.initmanager import InitManager
from boss.bo.kernel_factory import KernelFactory
from boss.bo.model import Model
from boss.bo.rstmanager import RstManager
from boss.io.main_output import MainOutput
import boss.io.parse as parse
from boss.settings import Settings, standardize_signature
from boss.utils.minimization import Minimization
from boss.utils.timer import Timer
from boss.bo.results import BOResults
import boss.keywords as bkw


class BOMain:
    """
    Class for handling Bayesian Optimization
    """

    def __init__(self, f, bounds, **keywords):
        keywords["bounds"] = bounds
        settings = Settings(keywords, f=f)
        self.settings = settings
        self.rst_manager = None
        self._setup()

    @classmethod
    def from_file(cls, ipfile, f=None, **new_keywords):
        """Initialize BOMain from a BOSS input or rst file.

        Parameters
        ----------
        ipfile : path_like
            The input file to initialize from, can be either a boss input or rst file.
        **new_keywords
            Any new BOSS keywords.
        """
        self = cls.__new__(cls)
        input_data = parse.parse_input_file(ipfile)
        rst_data = input_data.get("rst_data", np.array([]))
        keywords = input_data.get("keywords", {})
        keywords.update(new_keywords)
        self.settings = Settings(keywords, f=f)
        self.rst_manager = RstManager(self.settings, rst_data)
        cls._setup(self)
        return self

    @classmethod
    def from_settings(cls, settings, rst_data=None):
        """Construction from a Settings object. """
        self = cls.__new__(cls)
        self.settings = settings
        self.rst_manager = RstManager(self.settings, rst_data)
        cls._setup(self)
        return self

    def _setup(self):
        """Common setup for all factory methods. """
        settings = self.settings
        self.main_output = MainOutput(settings)
        if not self.rst_manager:
            self.rst_manager = RstManager(settings)

        self.init_manager = InitManager(
            settings["inittype"], settings["bounds"], settings["initpts"]
        )

        self.dim = settings.dim
        self.initpts = settings["initpts"]
        self.iterpts = settings["iterpts"]
        self.updatefreq = settings["updatefreq"]
        self.initupdate = settings["initupdate"]
        self.updateoffset = settings["updateoffset"]
        self.updaterestarts = settings["updaterestarts"]
        self.hmciters = settings["hmciters"]
        self.cores = settings["cores"]
        self.f = settings.f
        self.periods = settings["periods"]
        self.bounds = settings["bounds"]
        self.acqfn = settings.acqfn
        self.acqfnpars = settings["acqfnpars"]
        self.acqtol = settings["acqtol"]
        self.min_dist_acqs = settings["min_dist_acqs"]
        self.minzacc = settings["minzacc"]
        self.kerntype = settings["kernel"]
        self.noise = settings["noise"]
        self.ynorm = settings["ynorm"]
        self.timer = Timer()
        self.dir = settings.dir
        self.gm_tol = settings["gm_tol"]

        self.kernel = KernelFactory.construct_kernel(settings)
        self.model = None
        self.X_init = np.array([]).reshape(0, self.dim)
        self.Y_init = np.array([]).reshape(0, 1)
        self.convergence = np.array([]).reshape(0, 5)
        self.est_yrange = 0
        self.hmcsample = []
        self.exit_ok = False  # successful exit from main loop

        Minimization.set_parallel(self.settings["parallel_optims"])

    def get_x(self):
        """
        Returns the points where the objective function has been evaluated,
        in order of acquisition.
        """
        if self.model == None:
            return self.X_init
        else:
            return self.model.X

    def get_y(self):
        """
        Returns the evaluated values of the objective function, in order of
        acquisition. This method should always be called instead of
        BOMain.model.Y for any use outside of the BOMain class.
        """
        if self.model == None:
            return self.Y_init
        else:
            return self.model.Y

    def get_mu_and_nu(self, x):
        """
        Returns the model prediction (mu) and its standard deviation (nu)
        at point x.
        The standard deviation is calculated based on predictive variance
        with noise (model variance).
        """
        if self.model is None:
            return None, None
        else:
            mu, var = self.model.predict(x)
            nu = np.sqrt(var)
            return mu, nu

    def get_mu(self, x):
        """
        Returns the model prediction at point x.
        """
        return self.get_mu_and_nu(x)[0]

    def get_nu(self, x):
        """
        Returns the standard deviation of the model prediction at point x,
        calculated based on predictive variance with noise (model variance).
        """
        return self.get_mu_and_nu(x)[1]

    def get_grad(self, x):
        """
        Returns the predictive gradients.
        """
        g = self.model.predict_grads(x)
        return np.array([g[0], None])

    def _add_xy(self, x_new, y_new):
        """
        Internal functionality for saving a new acquisition (x, y).
        Initializes the GP model when the number ofacquisitions meets the
        pre-specified number of initialization points.
        """
        if self.model == None:
            self.X_init = np.vstack([self.X_init, x_new])
            self.Y_init = np.vstack([self.Y_init, y_new])
            if self.get_x().shape[0] == self.initpts:
                self._init_model()
        else:
            self.model.add_data(x_new, y_new)
            self.est_yrange = np.max(self.model.Y) - np.min(self.model.Y)

    def add_xy_list(self, x_new, y_new):
        """
        Saves multiple acquisitions. Initializes the GP model when the number
        of acquisitions meets the pre-specified number of initialization
        points.
        """
        for i in range(x_new.shape[0]):
            self._add_xy(x_new[i, :], y_new[i])

    def _init_model(self):
        """
        Initializes the GP model. This should be called when all initialization
        points have been evaluated. Further acquisition locations can then be
        queried by optimizing an acquisition function.
        """
        self.est_yrange = np.max(self.Y_init) - np.min(self.Y_init)
        self.model = Model(
            self.X_init,
            self.Y_init,
            self.kernel,
            self.noise,
            self.ynorm,
        )
        n = len(self.model.get_unfixed_params())
        self.unfixed_model_params = np.array([]).reshape(0, n)

    def _should_optimize(self, i_iter):
        """
        Returns True if the model should be optimized at iteration i.
        """
        bo_i = np.max([0, i_iter + 1 - self.initpts])  # == 0 means initial iters

        # No model -> no optimization
        if self.model == None:
            return False

        # Check if initialization has just completed and we want to optimize
        elif bo_i == 0:
            if self.initupdate:
                return True
            else:
                return False

        # Check if optimization threshold and interval have passed
        elif bo_i >= self.updateoffset and bo_i % self.updatefreq == 0:
            return True

        # Otherwise there is no need to optimize
        else:
            return False

    def run(self):
        """
        The Bayesian optimization main loop. Evaluates first the initialization
        points, then creates a GP model and uses it and an acquisition function
        to locate the next points where to evaluate. Stops when a pre-specified
        number of initialization points and BO points have been acquired or a
        convergence criterion is met.

        Returns
        -------
        BOResults
            An object that provides convenient access to the most
            important results from the optimization.
        """
        self.exit_ok = False

        # Resolve current iteration and set optimisation parameters
        i_current = len(self.get_x())
        if i_current == 0:
            self.main_output.new_file()
            self.rst_manager.new_file()

        initpts = self.settings["initpts"]
        iterpts = self.settings["iterpts"]
        minfreq = self.settings["minfreq"]

        totalpts = initpts + iterpts

        x_next = self._get_x_next(i_current)

        # BO main loop
        for i_iter in np.arange(i_current, totalpts):

            self.main_output.iteration_start(i_iter + 1)
            self.timer.startLap()

            # Evaluation
            x_new, y_new = self._evaluate(i_iter, x_next)

            # Store new data and refit model.
            #  - Create the model when all initial data has been acquired.
            for i in range(len(y_new)):
                self._add_xy(x_new[i], y_new[i])
                self.rst_manager.new_data(x_new[i], y_new[i])

            # Optimize model if needed.
            if self._should_optimize(i_iter):
                self._optimize_model(i_iter)

            # Find next acquisition point.
            x_next = self._get_x_next(i_iter + 1)

            # Add model parameters to rst-file.
            if self.model != None:
                mod_unfixed_par = self.model.get_unfixed_params()
                self.rst_manager.new_model_params(mod_unfixed_par)

                # Also store model parameters directly on BOMain
                self.unfixed_model_params = np.vstack(
                    (self.unfixed_model_params, mod_unfixed_par)
                )
            else:
                mod_unfixed_par = None

            # Convergence diagnostics, calculate iteration specific info
            iconv = i_iter - initpts + 1
            write_conv = iconv >= 0 and (
                iconv == iterpts or (minfreq > 0 and (iconv % minfreq == 0))
            )
            if write_conv:
                self._add_convergence()

            # Output to boss outfile
            # Print a summary of important iteration data
            self.main_output.iteration_summary(
                i_iter,
                self.get_y().size,
                x_new,
                y_new,
                self.convergence,
                x_next,
                self.est_yrange,
                mod_unfixed_par,
                self.timer,
            )

            # Stopping criterion: convergence
            if self._has_converged(i_iter):
                self.main_output.convergence_stop()
                break

        # If we get here, the optimization finished successfully.
        self.exit_ok = True
        res = self.get_results()
        return res

    def resume(self, iterpts):
        """Resumes an optimization.

        If the BOMain object still exists and has an initialized model,
        this method will continue where the previous optimization left off
        for a given number of new iterations. Existing out and rst files will
        be updated with data form the new iterations, but changes done to the settings upon
        resuming will not be reflected in the beginning of these files.

        Parameters
        ----------
        iterpts : int
            The number of iterations for the resumed optimization.
            This includes the number of iterations from the previous run,
            thus if you first ran for 10 iterations and now want 10 more,
            you should set iterpts=20.

        Returns
        -------
        BOResults
            A dict-like object that provides convenient access to the most
            important results from the optimization.
        """
        self.settings["iterpts"] = iterpts
        return self.run()

    def get_results(self):
        return BOResults(self)

    def _has_converged(self, i_iter):
        """
        Checks whether dxhat has been within tolerance for long enough
        TODO: should use dxmuhat instead?
        """
        if self.gm_tol is not None:
            conv_tol, conv_it = self.gm_tol
            if i_iter > self.initpts + conv_it:
                curr_xhat = self.convergence[-1, -3]
                within_tol = True
                for test_i in range(2, conv_it - 1):
                    if euclidean(self.convergence[-test_i, -3], curr_xhat) > conv_tol:
                        within_tol = False
                return within_tol
        return False

    def _get_x_next(self, i_iter):
        """
        Get a new point to evaluate by either reading it from the rst-file or,
        in case it doesn't contain the next point to evaluate, by obtaining
        a new initial point (when run is in initialization stage) or
        minimizing the acquisition function (when the run is in BO stage).
        """
        x_next = self.rst_manager.get_x(i_iter)
        if np.any(x_next == None):
            if i_iter < self.initpts:
                x_next = self.init_manager.get_x(i_iter)
            else:
                x_next, acqfn = self._acqnext(i_iter)
        return x_next

    def _acqnext(self, i_iter):
        """
        Selects the acquisition function to use and returns its x_next location
        as well as the used acquisition function.
        """
        if self.hmciters == 0:
            acqfn = self.acqfn
            expfn = explore
        else:
            hmc = HMCwrapper(self.hmcsample)
            acqfn = hmc.averageacq(self.acqfn, self.cores)
            expfn = hmc.averageacq(explore, self.cores)

        x_next = self._minimize_acqfn(acqfn)

        # check if x_next indicates we should trigger pure exploration
        if self._location_overconfident(x_next):
            x_next = self._minimize_acqfn(expfn)
            return x_next, expfn
        else:
            return x_next, acqfn

    def _minimize_acqfn(self, acqfn):
        """
        Minimizes the acquisition function to find the next
        sampling location 'x_next'.
        """
        # Calculate the number of local minimizers to start. This has two
        # steps:

        # 1. Estimate the number of local minima in the surrogate model.
        estimated_numpts = self.model.estimate_num_local_minima(self.bounds)

        # 2. Increase the estimate to approximate the number of local
        # minima in the acquisition function. Here we assume that the
        # increase is proportional to the estimated number of local minima
        # per dimension.
        minima_multiplier = 1.7
        estimated_numpts = (minima_multiplier ** self.dim) * estimated_numpts

        num_pts = min(len(self.model.X), int(estimated_numpts))

        # minimize acqfn to obtain sampling location
        gmin = Minimization.minimize_from_random(
            acqfn,
            self.bounds,
            num_pts=num_pts,
            shift=0.1 * np.array(self.periods),
            args=[self.model, self.acqfnpars],
        )
        return np.atleast_1d(np.array(gmin[0]))

    def _location_overconfident(self, x_next):
        """
        Checks is model variance is lower than tolerance at suggested x_next.
        """
        if self.acqtol is None:
            return False
        else:
            nu_x_next = np.sqrt(self.model.predict(x_next)[1])
            if nu_x_next >= self.acqtol:
                return False
            else:
                self.main_output.progress_msg(
                    "Acquisition location " + "too confident, doing pure exploration", 1
                )
                return True

    def _evaluate(self, i_iter, x_next):
        """
        Get a new evaluation either from the rst-file or, in case it doesn't
        contain the corresponding evaluation, by evaluating the user function
        """
        y_new = self.rst_manager.get_y(i_iter)
        if np.any(y_new == None):
            return self._evaluate_x_new(x_next)
        else:
            return np.atleast_2d(x_next), np.atleast_1d(y_new)

    def _evaluate_x_new(self, x_new):
        """
        Evaluates user function at given location 'x_new'
        to get the observation scalar 'y_new'.
        Later also gradient 'ydnew' should be made possible.
        """
        # run user script to evaluate sample
        self.main_output.progress_msg(
            "Evaluating objective function at x ="
            + self.main_output.utils.oneDarray_line(x_new.flatten(), self.dim, float),
            1,
        )

        local_timer = Timer()
        x_new = np.atleast_2d(x_new)  # X-data expected to be 2d-array
        y_new = np.atleast_1d(self.f(x_new))  # call the user function
        # os.chdir(self.dir)  # in case the user function changed the dir
        self.main_output.progress_msg(
            "Objective function evaluated,"
            + " time [s] %s" % (local_timer.str_lapTime()),
            1,
        )
        return x_new, y_new

    def _optimize_model(self, i_iter):
        """
        Optimize the GP model or, if the next hyperparameters are contained in
        the rst-file, just use them.
        """
        if self.hmciters > 0:
            self.hmcsample = self.model.sample_unfixed_params(self.hmciters)

        n = self.model.get_unfixed_params().size
        theta = self.rst_manager.get_theta(i_iter, n)
        if np.any(theta == None):
            self.model.optimize(self.updaterestarts)
        else:
            self.model.set_unfixed_params(theta)

    def _add_convergence(self):
        """
        Updates self.convergence with a new row containing
        bestx, besty, xhat, muhat, nuhat (in this order).
        """
        # if self.model == None:
        #     conv = np.atleast_2d(np.repeat(np.nan, 5))
        # else:
        bestx, besty = self.model.get_best_xy()
        xhat, muhat, nuhat = self._min_model_prediction()
        conv = np.atleast_2d([bestx, besty, xhat, muhat, nuhat])
        self.convergence = np.vstack((self.convergence, conv))

    def _min_model_prediction(self):
        """
        Finds and returns the model global minimum prediction xhat
        and the model mean (muhat) and variance (nuhat) at that point.
        """
        gmin = Minimization.minimize(
            self.model.predict_mean_grad,
            self.bounds,
            self.kerntype,
            np.hstack([self.model.X, self.model.Y]),
            self.min_dist_acqs,
            accuracy=self.minzacc,
        )
        xhat = np.array(gmin[0])
        muhat, nuhat = self.model.predict(np.atleast_2d(xhat))
        return xhat, muhat[0][0], nuhat[0][0]



"""
Functionality to output raw data on request to files outside of the main
output (*.out) file and the restart (*.rst) file.
"""
import numpy as np
import os

import boss.io.ioutils as ioutils
from boss.keywords import stringify


def dump_input_file(new_file_path, keywords, results_header=False):
    with open(new_file_path, "w") as fd:
        for key in sorted(keywords):
            val_str = stringify(keywords[key])
            fd.write(f"{key} {val_str}\n")
        if results_header:
            fd.write("\nRESULTS:")


def dump_model(settings, dest_file, bo, mod_params, xhat):
    """
    Outputs model slice (up to 2D) mean and variance in a grid to
    models/it#.dat
    """
    npts_per_dim = settings["pp_model_slice"][2]
    slice1 = settings["pp_model_slice"][0]
    slice2 = settings["pp_model_slice"][1]
    bounds = settings["bounds"]
    is_2d = slice1 != slice2

    # Construct query points for required slices
    x1 = np.linspace(*bounds[slice1], npts_per_dim)
    if is_2d:
        # 2D slice
        x2 = np.linspace(*bounds[slice2], npts_per_dim)
        x1, x2 = np.meshgrid(x1, x2)
        x1 = x1.ravel()
        x2 = x2.ravel()
        npts = npts_per_dim**2
    else:
        # 1D slice
        npts = npts_per_dim
    assert len(x1) == npts

    # Full ND array of query points
    X = np.empty((npts, settings.dim))

    # Fill ND query point array with default values (or xhat)
    if settings["pp_var_defaults"] is None:
        X[:] = xhat
    else:
        X[:] = settings["pp_var_defaults"]

    # Use actual query values at correct dimensions
    X[:, slice1] = x1
    if is_2d:
        X[:, slice2] = x2

    # Get data
    mu, nu = bo.get_mu_and_nu(X)
    model_data = np.hstack((X, mu, nu))

    # Build header lines
    header = "# Model output (x mu nu), grid of "
    if is_2d:
        header += "%ix%i=" % (npts_per_dim, npts_per_dim)
    header += "%i pts\n" % npts
    header += (
        "# Model parameter values "
        + "({}): ".format(' '.join(mod_params.keys()))
        + format(tuple(mod_params.values()))
    )

    # Write
    with open(dest_file, 'w') as fd:
        fd.write(header + '\n')
        np.savetxt(fd, model_data, fmt='%18.8e')


def dump_acqfn(settings, dest_file, bo, acqfn, defs):
    """
    Outputs acquisition function slice (up to 2D) in a grid to
    acqfns/it#.dat
    """
    acqf_data = [[]] * (settings.dim + 1)  # coords + acqf
    npts = settings["pp_model_slice"][2]
    coords = np.array(
        [
            np.linspace(settings["bounds"][i, 0], settings["bounds"][i, 1], npts)
            for i in settings["pp_model_slice"][:2]
        ]
    )
    defaults = (
        defs if settings["pp_var_defaults"] is None else settings["pp_var_defaults"]
    )
    if settings["pp_model_slice"][0] != settings["pp_model_slice"][1]:
        # 2D slice
        x1, x2 = np.meshgrid(coords[0], coords[1])
        for i in range(npts):
            for j in range(npts):
                p = np.array([x1[i, j], x2[i, j]])
                for d in range(settings.dim):
                    if d not in settings["pp_model_slice"][:2]:
                        p = np.insert(p, d, defaults[d])
                af, daf = acqfn(p, bo.model, bo.acqfnpars)
                p = np.insert(p, len(p), float(af))
                acqf_data = np.insert(acqf_data, len(acqf_data[0]), p, axis=1)
        titleLine = "# Acquisition function  (x" + " af), grid of %i pts" % (npts ** 2)
    else:
        # 1D slice
        x1 = coords[0]
        for i in range(npts):
            p = np.array([x1[i]])
            for d in range(settings.dim):
                if d not in settings["pp_model_slice"][:2]:
                    p = np.insert(p, d, defaults[d])
            af, daf = acqfn(p, bo.model, bo.acqfnpars)
            p = np.insert(p, len(p), float(af))
            acqf_data = np.insert(acqf_data, len(acqf_data[0]), p, axis=1)
        titleLine = "# Acquisition function (x" + " af), grid of %i pts" % (npts)

    ioutils.write_cols(dest_file, acqf_data, "    ", titleLine, "%18.8E")


def dump_truef(settings, dest_file, last_xhat):
    """
    Outputs true function slice (up to 2D) in a grid
    """
    truef_data = [[]] * (settings.dim + 1)  # coords + truef
    npts = settings["pp_truef_npts"]
    coords = np.array(
        [
            np.linspace(settings["bounds"][i, 0], settings["bounds"][i, 1], npts)
            for i in settings["pp_model_slice"][:2]
        ]
    )
    defaults = (
        last_xhat
        if settings["pp_var_defaults"] is None
        else settings["pp_var_defaults"]
    )
    if settings["pp_model_slice"][0] != settings["pp_model_slice"][1]:
        # 2D slice
        x1, x2 = np.meshgrid(coords[0], coords[1])
        for i in range(npts):
            for j in range(npts):
                p = np.array([x1[i, j], x2[i, j]])
                for d in range(settings.dim):
                    if d not in settings["pp_model_slice"][:2]:
                        p = np.insert(p, d, defaults[d])
                tf = settings.f(np.atleast_2d(p))
                os.chdir(settings.dir)
                p = np.insert(p, len(p), float(tf))
                truef_data = np.insert(truef_data, len(truef_data[0]), p, axis=1)
        titleLine = "# True function output (x" + " f), grid of %ix%i=%i pts" % (
            npts,
            npts,
            npts ** 2,
        )
    else:
        # 1D slice
        x1 = coords[0]
        for i in range(npts):
            p = np.array([x1[i]])
            for d in range(settings.dim):
                if d not in settings["pp_model_slice"][:2]:
                    p = np.insert(p, d, defaults[d])
            tf = settings.f(np.atleast_2d(p))
            os.chdir(settings.dir)
            p = np.insert(p, len(p), float(tf))
            truef_data = np.insert(truef_data, len(truef_data[0]), p, axis=1)
        titleLine = "# True function output (x" + " f), grid of %i pts" % (npts)

    ioutils.write_cols(dest_file, truef_data, "    ", titleLine, "%18.8E")


def dump_mep(path):
    """
    Outputs the coordinates of each minimum energy path into files
    """
    crds = path.crds
    i = path.mi
    j = path.mj
    s = "# From minima " + str(i) + " ("
    for crd in crds[0, :]:
        s += "  %f" % crd
    s += ")\n#   to minima " + str(j) + " ("
    for crd in crds[crds.shape[0] - 1, :]:
        s += "  %f" % crd
    s += ")\n"
    s += "# Highest energy: %8.3E\n\n" % path.maxe
    s += "# coordinates, energy\n"

    for ci in range(crds.shape[0]):
        for crd in crds[ci, :]:
            s += "  %f" % crd
        s += "  %8.3E\n" % path.energy[ci]
    ioutils.overwrite("mep/pathcrds_" + str(i) + "_" + str(j) + ".dat", s)



import numpy as np


class Path:
    def __init__(self, space, crds=None, i=None, j=None, energy=None):
        """
        connection  ...
        space       (Space) The Space class defines boundaries and periodicity.
        """
        self.mi = i
        self.mj = j
        self.space = space
        self.crds = crds
        self.energy = energy
        self.maxe = np.max(energy)

    def initFromConnection(self, connection):
        # Indexes of the minima that this path connects
        self.mi, self.mj = [connection[0].name, connection[1].name]

        # combine a path from the two trees
        end_node_A, end_node_B = connection
        parents_A = end_node_A.get_all_parents()
        parents_B = end_node_B.get_all_parents()
        parents_A.reverse()
        crds = np.array([]).reshape(0, end_node_A.coords.shape[0])
        for node in parents_A:
            crds = np.vstack((crds, node.coords))
        crds = np.vstack((crds, end_node_A.coords, end_node_B.coords))
        for node in parents_B:
            crds = np.vstack((crds, node.coords))
        self.crds = crds

    def reform(self, eps):
        """
        Reform with points at eps distance from another

        Parameters:
        eps         (float) Separation of points in the new path.
        """
        newpath = np.atleast_2d(self.crds[0, :])
        next_i = 1
        prev_pt = self.crds[0, :]  # point from where to find next
        direction = self.space.vec(self.crds[next_i, :], self.crds[next_i - 1, :])
        nextdist = np.linalg.norm(direction)  # dist to next point in orig path
        direction = direction / nextdist  # (make it a unit vector)
        prevdist = 0  # dist from prev point in new path
        while True:
            # Starting from first point in path, go along the path, adding
            # points at eps distance from another
            if nextdist >= (eps - prevdist):
                # A new point needs to be added before encountering the next
                # point in the old path
                newpoint = prev_pt + (eps - prevdist) * direction
                newpoint = self.space.crds(np.array([newpoint]))[0]
                newpath = np.vstack((newpath, newpoint))
                prev_pt = newpoint
                nextdist = nextdist - (eps - prevdist)
                prevdist = 0
            else:
                # Next point in old path encountered before a new point can be
                # added: change direction and remember how much distance was
                # covered before this direction change
                if next_i == self.crds.shape[0] - 1:  # if next point is the last
                    newpath = np.vstack((newpath, self.crds[next_i, :]))
                    self.crds = newpath
                    return
                d = np.linalg.norm(
                    self.space.vec(
                        newpath[newpath.shape[0] - 1, :], self.crds[next_i, :]
                    )
                )
                prevdist = d
                prev_pt = self.crds[next_i, :]
                next_i += 1
                direction = self.space.vec(
                    self.crds[next_i, :], self.crds[next_i - 1, :]
                )
                nextdist = np.linalg.norm(direction)
                direction = direction / nextdist



import datetime
import os

import numpy as np

import boss.io.ioutils as io
import boss.keywords as bkw
from boss import __version__


class MainOutput:
    """
    Functionality to write to the main output (*.out) file.
    """

    def __init__(self, settings):
        self.ipfile = settings["ipfile"]
        self.outfile = settings["outfile"]
        self.verbosity = settings["verbosity"]
        self.utils = io
        self.settings = settings

    def new_file(self):
        """Intializes a new main output file with header, input file and settings. """
        if os.path.isfile(self.outfile):
            print("warning: overwriting file '" + self.outfile + "'")

        self.header()
        if os.path.isfile(self.ipfile):
            self.ipfile_repeat(self.settings)

        self.add_settings(self.settings)

    def header(self):
        """
        Writes a header to main output file overwriting a possibly existing
        old output file at the same filepath.
        """
        s = (
            "\n-----------------------------   Welcome to ....   ----------"
            + "--------------------\n"
            + "                      _______  _______ _______ _______ \n"
            + "                     |   _   \|   _   |   _   |   _   |\n"
            + "                     |.  1   /|.  |   |   1___|   1___|\n"
            + "                     |.  _   \|.  |   |____   |____   |\n"
            + "                     |:  1    |:  1   |:  1   |:  1   |\n"
            + "                     |::.. .  |::.. . |::.. . |::.. . |\n"
            + "                     `-------'`-------`-------`-------'\n\n"
            + "{:^80s}\n".format("Version " + str(__version__))
            + "{:^80s}\n".format(datetime.datetime.now().strftime("%d-%m-%Y %H:%M:%S"))
            + "------------------------------------------------------------"
            + "--------------------\n\n"
        )
        with open(self.outfile, "w") as fd:
            fd.write(s)

    def footer(self, totaltime):
        """
        Writes a footer to main output file
        """
        text = "BOSS is done! Have a nice day :)"
        s = (
            "\n\n--------------------------------------------------------"
            + "------------------------\n"
        )
        s += "{:^80s}\n\n".format(text)
        s += "{:^40s}".format(
            datetime.datetime.now().strftime("Datetime %d-%m-%Y %H:%M:%S")
        )
        s += "{:^40s}".format("Total time [s] %s" % (totaltime))  # code total time
        s += (
            "\n---------------------------------------------------------"
            + "-----------------------\n"
        )
        with open(self.outfile, "a") as fd:
            fd.write(s)

    def ipfile_repeat(self, is_rst):
        """
        Repeats the input file near the beggining of the main output file.
        """
        if self.verbosity > 0:
            self.progress_msg("Reading BOSS input file from: " + self.ipfile, 0)
            self.progress_msg("Initializing...\n", 0)
            self.section_header("INPUT FILE")
            s = ""
            with open(self.ipfile, "r") as f:
                line = f.readline()
                while len(line) > 0:
                    s += line
                    line = f.readline()

            with open(self.outfile, "a") as fd:
                fd.write(s)

    def add_settings(self, settings):
        """
        Outputs the interpreted code variable settings to main output file.
        """
        if self.verbosity > 0:
            self.section_header("SIMULATION OPTIONS")
            s = "|| File input/output \n"
            s += "ipfile         %s\n" % (settings["ipfile"])
            s += "userfn         %s\n" % (bkw.stringify(settings["userfn"]))
            s += "outfile        %s\n" % (settings["outfile"])
            s += "rstfile        %s\n\n" % (settings["rstfile"])

            s += "|| Key settings \n"
            s += (
                "bounds        "
                + io.twoDfloatarray_line(settings["bounds"], settings.dim, 2)
                + "\n"
            )
            s += (
                "kerntype      "
                + io.oneDarray_line(settings["kernel"], settings.dim, str)
                + "\n"
            )
            if np.any(settings["kernel"] == "stdp"):
                s += (
                    "periods      "
                    + io.oneDarray_line(settings["periods"], settings.dim, float)
                    + "\n"
                )
            s += (
                "yrange        "
                + io.oneDarray_line(settings["yrange"], 2, float)
                + "\n"
            )
            s += "noise          %8.3E\n" % (settings["noise"])
            if not settings.is_rst:
                s += "inittype       %s\n" % (settings["inittype"])
            s += "initpts   %i    iterpts   %i\n" % (
                settings["initpts"],
                settings["iterpts"],
            )
            if settings["gm_tol"] is not None:
                s += "gm_tol         %8.3E   %i\n" % (
                    settings["gm_tol"][0],
                    settings["gm_tol"][1],
                )
            else:
                s += "gm_tol         none\n"
            s += "verbosity      %i\n\n" % (settings["verbosity"])

            s += "|| Data acquisition \n"
            s += "acqfn                %s\n" % (settings["acqfn_name"])
            s += "acqtol               "
            if settings["acqtol"] is None:
                s += "none\n"
            else:
                s += "%8.3E\n" % (settings["acqtol"])
            s += "\n"

            s += "|| GP hyperparameters \n"
            s += (
                "thetainit      "
                + io.oneDarray_line(settings["thetainit"], settings.dim, float)
                + "\n"
            )
            s += (
                "thetabounds    "
                + io.twoDfloatarray_line(settings["thetabounds"], settings.dim + 1, 2)
                + "\n"
            )
            s += "thetaprior      %s\n" % (settings["thetaprior"])
            s += (
                "thetapriorpar  "
                + io.twoDfloatarray_line(settings["thetapriorpar"], settings.dim + 1, 2)
                + "\n\n"
            )

            s += "|| Hyperparameter optimization\n"
            s += "updatefreq   %i  initupdate      %s\n" % (
                settings["updatefreq"],
                str(settings["initupdate"]),
            )
            s += "updateoffset %i  updaterestarts  %i\n\n" % (
                settings["updateoffset"],
                settings["updaterestarts"],
            )

            if len(settings["pp_iters"]) > 0:
                s += "| postprocessing\n"
                # s += "pp_iters          ="
                # s += io.oneDarray_line(settings["pp_iters"],len(settings["pp_iters"]),int)+"\n"
                s += "pp_models         = %s\n" % (settings["pp_models"])
                s += "pp_acq_funcs      = %s\n" % (settings["pp_acq_funcs"])
                s += "pp_truef_npts     = "
                if settings["pp_truef_npts"] is None:
                    s += "none\n"
                else:
                    s += "%i\n" % (settings["pp_truef_npts"])
                s += "pp_model_slice        ="
                s += (
                    io.oneDarray_line(
                        settings["pp_model_slice"], len(settings["pp_model_slice"]), int
                    )
                    + "\n"
                )
                if settings["pp_var_defaults"] is not None:
                    s += "pp_var_defaults   ="
                    s += (
                        io.oneDarray_line(
                            settings["pp_var_defaults"],
                            len(settings["pp_var_defaults"]),
                            float,
                        )
                        + "\n"
                    )
                s += "pp_truef_at_xhats = %s\n" % (settings["pp_truef_at_xhats"])
                s += "pp_local_minima   = "
                if settings["pp_local_minima"] is None:
                    s += "none\n\n"
                else:
                    s += "%i\n\n" % (settings["pp_local_minima"])
            with open(self.outfile, "a") as fd:
                fd.write(s)

    def progress_msg(self, msg, priority, preceding_bl=False, nospace=False):
        """
        Announce progress message to main output file depending on verbosity.
        """
        m = ""
        if self.verbosity >= priority:
            if preceding_bl:
                m = m + "\n"
            m = m + "|"
            if not nospace:
                m = m + " "
            m = m + msg + "\n"

        with open(self.outfile, "a") as fd:
            fd.write(m)

    def convergence_stop(self):
        """
        Announces BO stop due to global minimum convergence
        """
        msg = "Stopped BO due to global minimum prediction convergence"
        self.progress_msg(msg, 0, True)

    def iteration_start(self, i_iter):
        """
        Output section header for the new iteration.
        """
        initpts = self.settings["initpts"]
        if i_iter <= initpts:
            text = "INITIAL DATAPOINT " + str(i_iter)
        else:
            text = "BO ITERATION " + str(i_iter - initpts)
        s = (
            "\n--------------------------------------------------------"
            + "------------------------\n"
        )
        s += "{:^80}".format(text)
        s += (
            "\n---------------------------------------------------------"
            + "-----------------------\n"
        )
        with open(self.outfile, "a") as fd:
            fd.write(s)

    def iteration_summary(
        self,
        i_iter,
        datasize,
        newXs,
        newYs,
        convergence,
        xnext,
        est_yrange,
        model_params,
        timer,
    ):
        """
        Outputs info about one BO iteration to main output file
        """
        sts = self.settings
        initpts = sts["initpts"]
        iterpts = sts["iterpts"]
        totalpts = initpts + iterpts
        minfreq = sts["minfreq"]

        iconv = i_iter - initpts + 1
        write_conv = iconv >= 0 and (
            iconv == iterpts or (minfreq > 0 and (iconv % minfreq == 0))
        )

        if write_conv:
            xbest, ybest, xhat, muhat, nuhat = convergence[-1, :]
            d_xhat, d_muhat = np.nan, np.nan
            if convergence.shape[0] > 1:
                prev_xhat, prev_muhat = convergence[-2, [2, 3]]
                d_xhat = np.linalg.norm(xhat - prev_xhat)
                d_muhat = np.linalg.norm(muhat - prev_muhat) / est_yrange

        # self.progress_msg("-Iteration summary-", 0, True)
        s = "| Data point added to dataset (x y): \n"
        newXs = np.atleast_2d(newXs)
        newYs = np.atleast_1d(newYs)

        for i in range(len(newYs)):
            s += io.data_line(newXs[i], [newYs[i]], fstr="%18.10E")
        s += "\n| Total ensemble size: %i\n" % datasize

        # Model
        if not np.any(model_params == None):  # Only BO iterations
            if write_conv:
                if self.verbosity > 0:
                    s += "| Best acquisition, x_best y_best:\n"
                    s += io.data_line(xbest.flatten(), ybest.flatten(), fstr="%18.10E")
                s += "| Global minimum prediction, x_hat mu_hat +- nu_hat:\n"
                s += io.data_line(xhat.flatten(), [muhat, nuhat], fstr="%18.10E")
                # Convergence
                if self.verbosity > 1 and not np.any(np.isnan([d_xhat, d_muhat])):
                    s += "| Global minimum convergence, d_xhat d_muhat:\n"
                    s += io.data_line([d_xhat, d_muhat], fstr="%18.10E")

            if self.verbosity > 1:
                s += "\n| GP model hyperparameters:\n"
                s += io.data_line(model_params, fstr="%18.10E")
        if self.verbosity > 1:
            s += "| Next sampling location x_next:\n"
            s += io.data_line(xnext.flatten(), fstr="%18.10E")
        s += "\nIteration time [s]: %8.3f" % (timer.getLapTime())
        s += "        Total time [s]: %8.3f" % (timer.getTotalTime())
        with open(self.outfile, "a") as fd:
            fd.write(s + "\n")

    def section_header(self, text):
        """
        Writes a section header to main output file.
        """
        s = (
            "--------------------------------------------------------"
            + "------------------------\n"
        )
        s += "{:^80s}\n".format(text)
        s += (
            "---------------------------------------------------------"
            + "-----------------------\n"
        )
        with open(self.outfile, "a") as fd:
            fd.write(s)

    def mep_start(self, mep):
        """
        Writes MEP options and local minima
        """
        s = ""
        s += "|| MEP options\n"
        s += "precision %8d    maxE      %8.3E\n" % (mep.precision, mep.maxE)
        s += "rrtsteps  %8d    nebsteps   %8d\n" % (mep.rrtsteps, mep.nebsteps)
        s += "\n"
        s += "Energy threshold starting at %8.3E, stepsize %8.3E.\n" % (
            mep.e_start,
            mep.stepsize,
        )
        s += "\n"

        s += "|| Minima\n"
        s += "(pt index, coordinates)\n"
        for i in range(mep.min_points.shape[0]):
            s += str(i)
            for j in range(mep.min_points.shape[1]):
                s += " " + str(mep.min_points[i, j])
            s += "\n"
        s += "\n"

        with open(self.outfile, "a") as fd:
            fd.write(s)

    def mep_result(self, mep):
        """
        Writes the results of MEP
        """
        s = "\n"
        s += "|| MEP results\n"
        s += "(pt index, pt index, highest energy on the minimum energy path)\n"
        l = mep.min_points.shape[0]
        e = np.zeros((l, l))
        for path in mep.fullpaths:
            e[path.mi, path.mj] = path.maxe

        for i in range(l):
            for j in range(l):
                if i < j:
                    s += "%d %d %8.3E\n" % (i, j, e[i, j])

        s += "\n"
        with open(self.outfile, "a") as fd:
            fd.write(s)



from itertools import repeat
from multiprocessing import Pool


class HMCwrapper:
    def __init__(self, sample):
        self.sample = sample

    def averageacq(self, acqfunc, nprocs):
        def wrapped(x, model, params):
            """
            Evaluate acqfn with all params in HMC-sample using multiprocessing
            """
            # get acq values for the sample
            param = model.get_unfixed_params()
            f_acqu = np.atleast_2d(np.array([0.0]))
            scipygradient = np.atleast_2d(np.zeros(x.shape)).T
            chunks = [
                self.sample[
                    i::nprocs,
                ]
                for i in range(nprocs)
            ]
            pool = Pool(processes=nprocs)
            results = pool.map(
                self.g,
                zip(chunks, repeat(x), repeat(model), repeat(params), repeat(acqfunc)),
            )
            pool.close()
            for procres in results:
                for iterres in procres:
                    f, g = iterres
                    f_acqu += f
                    scipygradient += g
            # reset model to starting state
            model.set_unfixed_params(param)
            # average over the sample and return
            n = self.sample.shape[0]
            return f_acqu / n, np.asfortranarray(scipygradient / n)

        return wrapped

    def g(self, zipped):
        chunk, x, model, params, acqfunc = zipped
        res = []
        for p in chunk:
            model.set_unfixed_params(p)
            res.append(acqfunc(x, model, params))
        return res



import importlib
import inspect
import os
import sys
from pathlib import Path
import copy

import numpy as np

import boss.keywords as bkw
import boss.io.parse as parse
import boss.io.dump as dump
from boss.bo.acq.ei import ei
from boss.bo.acq.elcb import elcb
from boss.bo.acq.exploit import exploit
from boss.bo.acq.explore import explore
from boss.bo.acq.lcb import lcb
from boss.io.main_output import MainOutput
from boss.utils.distributions import Distributions


def standardize_signature(f, xdim):
    """Standardizes the signature of the user-function.

    Internally, 2d np.ndarray are passed to the user function,
    thus user-functions with different signatures must first
    be converted to this format.

    Parameters
    ----------
    f: Callable
        User-function that takes an arbitary number
        of type-identical arguments,
    xdim: int
        The dimension, as returned by np.ndim, of the individual arguments to f.

    Returns
    -------
    Callable
        A wrapped (if necessary) version of f that takes a single
        1d np.ndarray as input.
    """
    sig = inspect.signature(f)
    nargs = len(sig.parameters)
    if xdim == 0:
        def fnew(x):
            return f(*x[0])
    elif xdim == 1:
        if nargs == 1:
            def fnew(x):
                return f(x[0])
        else:
            def fnew(x):
                return f(*np.hsplit(x[0], nargs))
    elif xdim == 2:
        if nargs == 1:
            fnew = f
        else:
            def fnew(x):
                return f(*np.hsplit(x, nargs))
    else:
        raise ValueError
    return fnew


class Settings(dict):
    """Reads, interprets and defines the code internal settings based on input.
    """

    def __init__(self, keywords, f=None):
        super().__init__()

        # Non-keyword attributes.
        self.is_rst = False

        self.dir = os.getcwd()
        self.acqfn = None

        # Before any other keywords are set, we assign values to all keywords
        # with independent defaults.
        self.set_independent_defaults(only_missing=False)

        # Update with keywords from config files
        self.update(parse.parse_config())

        # Update with BOSS keywords passed to init
        self.update(keywords)
        if keywords.get("bounds") is None:
            raise ValueError(
                "Keyword 'bounds' has to be defined by the user"
            )

        # Handle the user function: if a function is passed directly we take that,
        # otherwise fall back to the function specified by the userfn keyword.
        if f is None:
            if self["userfn"] is not None:
                f_user = bkw.func_from_keyword(self["userfn"])
                f = standardize_signature(f_user, self["userfn_arrdim"])
            self.f = f
        else:
            self["userfn"] = bkw.func_to_keyword(f)
            self.f = standardize_signature(f, self["userfn_arrdim"])

        # Set default values for dependent keywords if they
        # have not yet been specified.
        self.set_acqfn(self["acqfn_name"])
        self.correct()
        self.set_dependent_defaults(only_missing=True)

        # Set RNG seed if specified.
        # TODO: Propagate this seed to GPy to eliminate all randomness.
        if self["seed"] is not None:
            np.random.seed(self["seed"])

    @classmethod
    def from_file(cls, file_path):
        """Factory method for Constructing a Settings object from a boss input file.

        Parameters
        -----------
        file_path: str, Path
            Path to the input file.

        Returns
        -------
        Settings
            Settings object generated using the input file.
        """
        input_data = parse.parse_input_file(file_path, skip="results")
        settings = cls(input_data['keywords'])
        settings.is_rst = input_data['is_rst']
        return settings

    def copy(self):
        return copy.deepcopy(self)

    def set_independent_defaults(self, only_missing=True):
        """Sets default values for independent keywords. """
        if not only_missing:
            for cat in bkw.get_copied_categories():
                self.update(cat)
        else:
            for cat in bkw.get_copied_categories():
                for key, val in cat.items():
                    if self.get(key) is None:
                        self[key] = val

    def set_dependent_defaults(self, only_missing=True):
        """Sets default values for keywords that depend on other keywords.
        """
        should_update = lambda key: self.get(key) is None or only_missing is False

        if should_update("periods"):
            self["periods"] = self["bounds"][:, 1] - self["bounds"][:, 0]
        if should_update("iterpts"):
            self["iterpts"] = int(15 * self.dim ** 1.5)
        if should_update("pp_iters"):
            self["pp_iters"] = list(range(self["iterpts"] + 1))
            self["pp_iters"] = np.arange(0, self['iterpts'] + 1)
        if should_update("min_dist_acqs"):
            self["min_dist_acqs"] = 0.01 * min(self["periods"])

        # Default initial hyperparameters.
        if should_update("thetainit"):
            if self["ynorm"]:
                diff = 1.0 # when observations are range normalised
            else:
                diff = self["yrange"][1] - self["yrange"][0]
            self["thetainit"] = [0.5 * diff]  # sig
            for i in range(self.dim):  # lengthscales
                if self["kernel"][i] == "stdp":  # pbc
                    self["thetainit"].append(np.pi / 10)
                else:  # nonpbc
                    self["thetainit"].append(self["periods"][i] / 20)

        # Default hyperparameter constraints.
        if self["thetaprior"] is not None:
            if should_update("thetabounds"):
                self["thetabounds"] = [
                    [self["thetainit"][0] / 1000.0, self["thetainit"][0] * 1000.0]
                ]  # variance
                for i in range(self.dim):  # lengthscale
                    self["thetabounds"].append(
                        [
                            self["thetainit"][i + 1] / 100.0,
                            self["thetainit"][i + 1] * 100.0,
                        ]
                    )
                self["thetabounds"] = np.array(self["thetabounds"])

            # Default hyperparameter priors.
            if should_update("thetapriorpar"):
                if self["ynorm"]:
                    diff = 1.0 # when observations are range normalised
                else:
                    diff = self["yrange"][1] - self["yrange"][0]    
                if self["thetaprior"] == "gamma":

                    # Ulpu's heuristic prior
                    shape = 2.00
                    rate = 2.0/(diff/2.0)**2

                    # Original solution, to be tested further
                    #shape, rate = Distributions.gammaparams(
                    #    (diff/4)**2, (10*diff/4)**2, 0.5, 0.99)
                    #shape = 1.0    # NORMALIZATION
                    #rate = 1.5     # NORMALIZATION
                    self["thetapriorpar"] = [[shape, rate]]

                    for i in range(self.dim):
                        if self["kernel"][i] == "stdp":  # pbc
                            shape = 3.3678
                            rate = 9.0204
                        else:  # nonpbc
                            shape, rate = Distributions.gammaparams(
                                self["periods"][i] / 20, self["periods"][i] / 2
                            )
                        self["thetapriorpar"].append([shape, rate])
                    self["thetapriorpar"] = np.asarray(self["thetapriorpar"])
                else:
                    raise TypeError(
                        f"Unknown options set for thetaprior: {self['thetaprior']}."
                    )

        # Model slice and number of points.
        if should_update("pp_model_slice"):
            if self.dim == 1:
                self["pp_model_slice"] = np.array([1, 1, 50])
            else:
                self["pp_model_slice"] = np.array([1, 2, 25])

    @property
    def dim(self):
        """The dimensionality of the user-supplied objective.

        The number of dimensions is a read-only propery that is
        derived from the bounds provided by the user.

        Returns
        -------
        int
            The dimensionality of the objective.

        """
        return len(self["bounds"])

    @dim.setter
    def dim(self, val):
        raise AttributeError("Cannot modify read-only attribute: dim")

    def set_acqfn(self, name):
        if name == "elcb":
            self.acqfn = elcb
        elif name == "lcb":
            self.acqfn = lcb
            if len(self["acqfnpars"]) < 1:
                self["acqfnpars"] = np.array([2.0])  # explr_weight
        elif name == "explore":
            self.acqfn = explore
        elif name == "exploit":
            self.acqfn = exploit
        elif name == "ei":
            self.acqfn = ei
        else:
            raise TypeError(f"ERROR: Unknown acquisition function selected: {name}")

    def dump(self, file_path):
        """Writes the current settings to a boss input file.

        Parameters
        ----------
        fname : Union[str, path]
            Path to the destination file.
        """
        dump.dump_input_file(file_path, self)

    def correct(self):
        """Corrects the type and value of certain keywords.

        The user is allowed to be lazy in defining certain keywords,
        e.g., by providing lists instead of np.arrays, and we correct
        for this laziness here.
        """
        if self["pp_acq_funcs"] and self["verbosity"] < 2:
            self["verbosity"] = 2  # to have xnexts printed

        # Make sure int and float arrays are np and not python sequences.
        # and have the correct dtype
        for key, val in self.items():
            cat = bkw.find_category(key)
            cat_type, cat_dim = cat[0], cat[1]

            if cat_dim > 0 and cat_type != str and val is not None:
                if cat_type == int:
                    self[key] = np.asarray(val, dtype=np.int64)
                elif cat_type == float:
                    self[key] = np.asarray(val, dtype=np.float64)

        bounds = self['bounds']
        if bounds.ndim == 1:
            self['bounds'] = np.atleast_2d(bounds)

        kernel = self["kernel"]
        if isinstance(kernel, str):
            self["kernel"] = [kernel]

        if len(self["kernel"]) == 1:
            self["kernel"] *= self.dim

        # Exception to the above rule: gm_tol is a mixed list [float, int]
        # TODO: break gm_tol into two keywords in a future release
        gm_tol = self["gm_tol"]
        if gm_tol is not None:
            self["gm_tol"] = [float(gm_tol[0]), int(gm_tol[1])]



import numpy as np


def spring_force(points, i, k, space):
    v1 = space.vec(points[i - 1, :], points[i, :])
    v2 = space.vec(points[i + 1, :], points[i, :])
    return k * (v1 + v2)


def component_in_line(vector, direction):
    return direction * sum(vector * direction) / np.linalg.norm(direction) ** 2


def get_neb_tuning(points, bo, space, maxf):
    # if a point is moved directly towards its neighbour by the mean
    # distance between points, it will get assigned maximum parallel force
    dist = np.linalg.norm(space.vec(points[0, :], points[1, :]))
    spring_k = maxf / (2 * dist)

    # maximum perpendicular force is assigned to the steepest gradient
    n_pts = points.shape[0]
    dim = points.shape[1]
    gradients = bo.get_grad(points)[0]
    maxg = 0
    for i in range(n_pts):
        temp = np.zeros(dim)
        for j in range(dim):
            temp[j] = gradients[i][j][0]
        length = np.linalg.norm(temp)
        if length > maxg:
            maxg = length
    grad_k = maxf / maxg
    return spring_k, grad_k


def neb(points, bo, space, maxiter=1000):
    # initialize
    n_pts = points.shape[0]
    dim = points.shape[1]
    maxf = np.min(space.bounds[1, :] - space.bounds[0, :]) / 100
    spring_k, grad_k = get_neb_tuning(points, bo, space, maxf)

    for abc in range(maxiter):
        # get gradients
        gradients = bo.get_grad(points)[0]
        temp = np.zeros(n_pts * dim).reshape(n_pts, dim)
        for i in range(n_pts):
            for j in range(dim):
                temp[i, j] = gradients[i][j][0]
        gradients = temp

        # get spring forces
        spring_f = np.zeros(n_pts * dim).reshape(n_pts, dim)
        for i in range(1, n_pts - 1):
            spring_f[i] = spring_force(points, i, spring_k, space)

        # get and combine components
        forces = np.zeros(n_pts * dim).reshape(n_pts, dim)
        worst = 0.0
        for i in range(1, n_pts - 1):
            parallel = space.vec(points[i - 1, :], points[i + 1, :])
            parallel_f = component_in_line(spring_f[i], parallel)
            perpendic = -gradients[i] * grad_k
            perpendic_f = perpendic - component_in_line(perpendic, parallel)
            forces[i, :] = parallel_f + perpendic_f
            fnorm = np.linalg.norm(forces[i, :])
            worst = np.max((worst, fnorm))
            if fnorm > maxf:
                forces[i, :] = forces[i, :] / fnorm * maxf

        prev_pts = points.copy()
        points = points + forces
        points = space.crds(points)
        outofbounds = np.isnan(points[:, 0])
        points[outofbounds, :] = prev_pts[outofbounds, :]

        if worst < 0.1 * maxf:
            break

    return points



#!/usr/bin/env python

import math

from numpy import *


def i4_bit_hi1(n):

    # *****************************************************************************80
    #
    ## I4_BIT_HI1 returns the position of the high 1 bit base 2 in an integer.
    #
    #  Example:
    #
    #       N    Binary     BIT
    #    ----    --------  ----
    #       0           0     0
    #       1           1     1
    #       2          10     2
    #       3          11     2
    #       4         100     3
    #       5         101     3
    #       6         110     3
    #       7         111     3
    #       8        1000     4
    #       9        1001     4
    #      10        1010     4
    #      11        1011     4
    #      12        1100     4
    #      13        1101     4
    #      14        1110     4
    #      15        1111     4
    #      16       10000     5
    #      17       10001     5
    #    1023  1111111111    10
    #    1024 10000000000    11
    #    1025 10000000001    11
    #
    #  Licensing:
    #
    #    This code is distributed under the MIT license.
    #
    #  Modified:
    #
    #    22 February 2011
    #
    #  Author:
    #
    #    Original MATLAB version by John Burkardt.
    #    PYTHON version by Corrado Chisari
    #
    #  Parameters:
    #
    #    Input, integer N, the integer to be measured.
    #    N should be nonnegative.  If N is nonpositive, the value will always be 0.
    #
    #    Output, integer BIT, the number of bits base 2.
    #
    i = int(n)
    bit = 0
    while True:
        if i <= 0:
            break
        bit += 1
        i = i // 2
    return bit


def i4_bit_lo0(n):

    # *****************************************************************************80
    #
    ## I4_BIT_LO0 returns the position of the low 0 bit base 2 in an integer.
    #
    #  Example:
    #
    #       N    Binary     BIT
    #    ----    --------  ----
    #       0           0     1
    #       1           1     2
    #       2          10     1
    #       3          11     3
    #       4         100     1
    #       5         101     2
    #       6         110     1
    #       7         111     4
    #       8        1000     1
    #       9        1001     2
    #      10        1010     1
    #      11        1011     3
    #      12        1100     1
    #      13        1101     2
    #      14        1110     1
    #      15        1111     5
    #      16       10000     1
    #      17       10001     2
    #    1023  1111111111    11
    #    1024 10000000000     1
    #    1025 10000000001     2
    #
    #  Licensing:
    #
    #    This code is distributed under the MIT license.
    #
    #  Modified:
    #
    #    08 February 2018
    #
    #  Author:
    #
    #    Original MATLAB version by John Burkardt.
    #    Python version by Corrado Chisari
    #
    #  Parameters:
    #
    #    Input, integer N, the integer to be measured.
    #    N should be nonnegative.
    #
    #    Output, integer BIT, the position of the low 1 bit.
    #
    bit = 0
    i = int(n)
    while 1:
        bit = bit + 1
        i2 = i // 2
        if i == 2 * i2:
            break
        i = i2

    return bit


def i4_sobol_generate(m, n, skip):

    # *****************************************************************************80
    #
    ## I4_SOBOL_GENERATE generates a Sobol dataset.
    #
    #  Licensing:
    #
    #    This code is distributed under the MIT license.
    #
    #  Modified:
    #
    #    22 February 2011
    #
    #  Author:
    #
    #    Original MATLAB version by John Burkardt.
    #    PYTHON version by Corrado Chisari
    #
    #  Parameters:
    #
    #    Input, integer M, the spatial dimension.
    #
    #    Input, integer N, the number of points to generate.
    #
    #    Input, integer SKIP, the number of initial points to skip.
    #
    #    Output, real R(M,N), the points.
    #
    r = zeros((m, n))
    for j in range(1, n + 1):
        seed = skip + j - 2
        [r[0:m, j - 1], seed] = i4_sobol(m, seed)
    return r


def i4_sobol(dim_num, seed):

    # *****************************************************************************80
    #
    ## I4_SOBOL generates a new quasirandom Sobol vector with each call.
    #
    #  Discussion:
    #
    #    The routine adapts the ideas of Antonov and Saleev.
    #
    #  Licensing:
    #
    #    This code is distributed under the MIT license.
    #
    #  Modified:
    #
    #    22 February 2011
    #
    #  Author:
    #
    #    Original FORTRAN77 version by Bennett Fox.
    #    MATLAB version by John Burkardt.
    #    PYTHON version by Corrado Chisari
    #
    #  Reference:
    #
    #    Antonov, Saleev,
    #    USSR Computational Mathematics and Mathematical Physics,
    #    olume 19, 1980, pages 252 - 256.
    #
    #    Paul Bratley, Bennett Fox,
    #    Algorithm 659:
    #    Implementing Sobol's Quasirandom Sequence Generator,
    #    ACM Transactions on Mathematical Software,
    #    Volume 14, Number 1, pages 88-100, 1988.
    #
    #    Bennett Fox,
    #    Algorithm 647:
    #    Implementation and Relative Efficiency of Quasirandom
    #    Sequence Generators,
    #    ACM Transactions on Mathematical Software,
    #    Volume 12, Number 4, pages 362-376, 1986.
    #
    #    Ilya Sobol,
    #    USSR Computational Mathematics and Mathematical Physics,
    #    Volume 16, pages 236-242, 1977.
    #
    #    Ilya Sobol, Levitan,
    #    The Production of Points Uniformly Distributed in a Multidimensional
    #    Cube (in Russian),
    #    Preprint IPM Akad. Nauk SSSR,
    #    Number 40, Moscow 1976.
    #
    #  Parameters:
    #
    #    Input, integer DIM_NUM, the number of spatial dimensions.
    #    DIM_NUM must satisfy 1 <= DIM_NUM <= 40.
    #
    #    Input/output, integer SEED, the "seed" for the sequence.
    #    This is essentially the index in the sequence of the quasirandom
    #    value to be generated. On output, SEED has been set to the
    #    appropriate next value, usually simply SEED+1.
    #    If SEED is less than 0 on input, it is treated as though it were 0.
    #    An input value of 0 requests the first (0-th) element of the sequence.
    #
    #    Output, real QUASI(DIM_NUM), the next quasirandom vector.
    #
    global atmost
    global dim_max
    global dim_num_save
    global initialized
    global lastq
    global log_max
    global maxcol
    global poly
    global recipd
    global seed_save
    global v

    if not "initialized" in globals().keys():
        initialized = 0
        dim_num_save = -1

    if not initialized or dim_num != dim_num_save:
        initialized = 1
        dim_max = 40
        dim_num_save = -1
        log_max = 30
        seed_save = -1
        #
        #   Initialize (part of) V.
        #
        v = zeros((dim_max, log_max))
        v[0:40, 0] = transpose(
            [
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
                1,
            ]
        )

        v[2:40, 1] = transpose(
            [
                1,
                3,
                1,
                3,
                1,
                3,
                3,
                1,
                3,
                1,
                3,
                1,
                3,
                1,
                1,
                3,
                1,
                3,
                1,
                3,
                1,
                3,
                3,
                1,
                3,
                1,
                3,
                1,
                3,
                1,
                1,
                3,
                1,
                3,
                1,
                3,
                1,
                3,
            ]
        )

        v[3:40, 2] = transpose(
            [
                7,
                5,
                1,
                3,
                3,
                7,
                5,
                5,
                7,
                7,
                1,
                3,
                3,
                7,
                5,
                1,
                1,
                5,
                3,
                3,
                1,
                7,
                5,
                1,
                3,
                3,
                7,
                5,
                1,
                1,
                5,
                7,
                7,
                5,
                1,
                3,
                3,
            ]
        )

        v[5:40, 3] = transpose(
            [
                1,
                7,
                9,
                13,
                11,
                1,
                3,
                7,
                9,
                5,
                13,
                13,
                11,
                3,
                15,
                5,
                3,
                15,
                7,
                9,
                13,
                9,
                1,
                11,
                7,
                5,
                15,
                1,
                15,
                11,
                5,
                3,
                1,
                7,
                9,
            ]
        )

        v[7:40, 4] = transpose(
            [
                9,
                3,
                27,
                15,
                29,
                21,
                23,
                19,
                11,
                25,
                7,
                13,
                17,
                1,
                25,
                29,
                3,
                31,
                11,
                5,
                23,
                27,
                19,
                21,
                5,
                1,
                17,
                13,
                7,
                15,
                9,
                31,
                9,
            ]
        )

        v[13:40, 5] = transpose(
            [
                37,
                33,
                7,
                5,
                11,
                39,
                63,
                27,
                17,
                15,
                23,
                29,
                3,
                21,
                13,
                31,
                25,
                9,
                49,
                33,
                19,
                29,
                11,
                19,
                27,
                15,
                25,
            ]
        )

        v[19:40, 6] = transpose(
            [
                13,
                33,
                115,
                41,
                79,
                17,
                29,
                119,
                75,
                73,
                105,
                7,
                59,
                65,
                21,
                3,
                113,
                61,
                89,
                45,
                107,
            ]
        )

        v[37:40, 7] = transpose([7, 23, 39])
        #
        #   Set POLY.
        #
        poly = [
            1,
            3,
            7,
            11,
            13,
            19,
            25,
            37,
            59,
            47,
            61,
            55,
            41,
            67,
            97,
            91,
            109,
            103,
            115,
            131,
            193,
            137,
            145,
            143,
            241,
            157,
            185,
            167,
            229,
            171,
            213,
            191,
            253,
            203,
            211,
            239,
            247,
            285,
            369,
            299,
        ]

        atmost = 2 ** log_max - 1
        #
        #   Find the number of bits in ATMOST.
        #
        maxcol = i4_bit_hi1(atmost)
        #
        #   Initialize row 1 of V.
        #
        v[0, 0:maxcol] = 1

    #
    #   Things to do only if the dimension changed.
    #
    if dim_num != dim_num_save:
        #
        #   Check parameters.
        #
        if dim_num < 1 or dim_max < dim_num:
            print("I4_SOBOL - Fatal error!")
            print("   The spatial dimension DIM_NUM should satisfy:")
            print("       1 <= DIM_NUM <= %d" % dim_max)
            print("   But this input value is DIM_NUM = %d" % dim_num)
            return

        dim_num_save = dim_num
        #
        #   Initialize the remaining rows of V.
        #
        for i in range(2, dim_num + 1):
            #
            #   The bits of the integer POLY(I) gives the form of polynomial I.
            #
            #   Find the degree of polynomial I from binary encoding.
            #
            j = poly[i - 1]
            m = 0
            while 1:
                j = math.floor(j / 2.0)
                if j <= 0:
                    break
                m = m + 1
            #
            #   Expand this bit pattern to separate components of the logical array INCLUD.
            #
            j = poly[i - 1]
            includ = zeros(m)
            for k in range(m, 0, -1):
                j2 = math.floor(j / 2.0)
                includ[k - 1] = j != 2 * j2
                j = j2
            #
            #   Calculate the remaining elements of row I as explained
            #   in Bratley and Fox, section 2.
            #
            for j in range(m + 1, maxcol + 1):
                newv = v[i - 1, j - m - 1]
                l = 1
                for k in range(1, m + 1):
                    l = 2 * l
                    if includ[k - 1]:
                        newv = bitwise_xor(int(newv), int(l * v[i - 1, j - k - 1]))
                v[i - 1, j - 1] = newv
        #
        #   Multiply columns of V by appropriate power of 2.
        #
        l = 1
        for j in range(maxcol - 1, 0, -1):
            l = 2 * l
            v[0:dim_num, j - 1] = v[0:dim_num, j - 1] * l
        #
        #   RECIPD is 1/(common denominator of the elements in V).
        #
        recipd = 1.0 / (2 * l)
        lastq = zeros(dim_num)

    seed = int(math.floor(seed))

    if seed < 0:
        seed = 0

    if seed == 0:
        l = 1
        lastq = zeros(dim_num)

    elif seed == seed_save + 1:
        #
        #   Find the position of the right-hand zero in SEED.
        #
        l = i4_bit_lo0(seed)

    elif seed <= seed_save:

        seed_save = 0
        l = 1
        lastq = zeros(dim_num)

        for seed_temp in range(int(seed_save), int(seed)):
            l = i4_bit_lo0(seed_temp)
            for i in range(1, dim_num + 1):
                lastq[i - 1] = bitwise_xor(int(lastq[i - 1]), int(v[i - 1, l - 1]))

        l = i4_bit_lo0(seed)

    elif seed_save + 1 < seed:

        for seed_temp in range(int(seed_save + 1), int(seed)):
            l = i4_bit_lo0(seed_temp)
            for i in range(1, dim_num + 1):
                lastq[i - 1] = bitwise_xor(int(lastq[i - 1]), int(v[i - 1, l - 1]))

        l = i4_bit_lo0(seed)
    #
    #   Check that the user is not calling too many times!
    #
    if maxcol < l:
        print("I4_SOBOL - Fatal error!")
        print("   Too many calls!")
        print("   MAXCOL = %d\n" % maxcol)
        print("   L =         %d\n" % l)
        return
    #
    #   Calculate the new components of QUASI.
    #
    quasi = zeros(dim_num)
    for i in range(1, dim_num + 1):
        quasi[i - 1] = lastq[i - 1] * recipd
        lastq[i - 1] = bitwise_xor(int(lastq[i - 1]), int(v[i - 1, l - 1]))

    seed_save = seed
    seed = seed + 1

    return [quasi, seed]


def i4_uniform(a, b, seed):

    # *****************************************************************************80
    #
    ## I4_UNIFORM returns a scaled pseudorandom I4.
    #
    #  Discussion:
    #
    #    The pseudorandom number will be scaled to be uniformly distributed
    #    between A and B.
    #
    #  Licensing:
    #
    #    This code is distributed under the MIT license.
    #
    #  Modified:
    #
    #    22 February 2011
    #
    #  Author:
    #
    #    Original MATLAB version by John Burkardt.
    #    PYTHON version by Corrado Chisari
    #
    #  Reference:
    #
    #    Paul Bratley, Bennett Fox, Linus Schrage,
    #    A Guide to Simulation,
    #    Springer Verlag, pages 201-202, 1983.
    #
    #    Pierre L'Ecuyer,
    #    Random Number Generation,
    #    in Handbook of Simulation,
    #    edited by Jerry Banks,
    #    Wiley Interscience, page 95, 1998.
    #
    #    Bennett Fox,
    #    Algorithm 647:
    #    Implementation and Relative Efficiency of Quasirandom
    #    Sequence Generators,
    #    ACM Transactions on Mathematical Software,
    #    Volume 12, Number 4, pages 362-376, 1986.
    #
    #    Peter Lewis, Allen Goodman, James Miller
    #    A Pseudo-Random Number Generator for the System/360,
    #    IBM Systems Journal,
    #    Volume 8, pages 136-143, 1969.
    #
    #  Parameters:
    #
    #    Input, integer A, B, the minimum and maximum acceptable values.
    #
    #    Input, integer SEED, a seed for the random number generator.
    #
    #    Output, integer C, the randomly chosen integer.
    #
    #    Output, integer SEED, the updated seed.
    #
    if seed == 0:
        print("I4_UNIFORM - Fatal error!")
        print("   Input SEED = 0!")

    seed = math.floor(seed)
    a = round(a)
    b = round(b)

    seed = mod(seed, 2147483647)

    if seed < 0:
        seed = seed + 2147483647

    k = math.floor(seed / 127773)

    seed = 16807 * (seed - k * 127773) - k * 2836

    if seed < 0:
        seed = seed + 2147483647

    r = seed * 4.656612875e-10
    #
    #   Scale R to lie between A-0.5 and B+0.5.
    #
    r = (1.0 - r) * (min(a, b) - 0.5) + r * (max(a, b) + 0.5)
    #
    #   Use rounding to convert R to an integer between A and B.
    #
    value = round(r)

    value = max(value, min(a, b))
    value = min(value, max(a, b))

    c = value

    return [int(c), int(seed)]


def prime_ge(n):

    # *****************************************************************************80
    #
    ## PRIME_GE returns the smallest prime greater than or equal to N.
    #
    #  Example:
    #
    #      N    PRIME_GE
    #
    #    -10     2
    #      1     2
    #      2     2
    #      3     3
    #      4     5
    #      5     5
    #      6     7
    #      7     7
    #      8    11
    #      9    11
    #     10    11
    #
    #  Licensing:
    #
    #    This code is distributed under the MIT license.
    #
    #  Modified:
    #
    #    22 February 2011
    #
    #  Author:
    #
    #    Original MATLAB version by John Burkardt.
    #    PYTHON version by Corrado Chisari
    #
    #  Parameters:
    #
    #    Input, integer N, the number to be bounded.
    #
    #    Output, integer P, the smallest prime number that is greater
    #    than or equal to N.
    #
    p = max(math.ceil(n), 2)
    while not isprime(p):
        p = p + 1

    return p


def isprime(n):

    # *****************************************************************************80
    #
    ## IS_PRIME returns True if N is a prime number, False otherwise
    #
    #  Licensing:
    #
    #    This code is distributed under the MIT license.
    #
    #  Modified:
    #
    #    22 February 2011
    #
    #  Author:
    #
    #    Corrado Chisari
    #
    #  Parameters:
    #
    #    Input, integer N, the number to be checked.
    #
    #    Output, boolean value, True or False
    #
    if n != int(n) or n < 1:
        return False
    p = 2
    while p < n:
        if n % p == 0:
            return False
        p += 1

    return True



import numpy as np
import scipy.stats


def ei(x, model, params):
    """
    Expected improvement acquisition function

    Doesn't take any parameters (apart from the model).
    """

    minacq = np.min(model.Y)
    m, v = model.predict(np.atleast_2d(x), noise=False)
    s = np.sqrt(v)
    dmdx, dvdx = model.predict_grads(np.atleast_2d(x))
    dmdx = dmdx[:, :, 0]
    z = (minacq - m) / s
    phi = scipy.stats.norm.pdf(z)
    Phi = scipy.stats.norm.cdf(z)
    f_acqu = -s * (z * Phi + phi)
    df_acqu = dmdx * Phi - dvdx / (2 * s) * phi
    scipygradient = np.asmatrix(df_acqu).transpose()
    return f_acqu, scipygradient



import numpy as np
import warnings
import itertools

import boss.utils.sobol_seq as sobol_seq


class InitManager:
    def __init__(self, inittype, bounds, initpts):
        """
        Creates initial points which can be queried with the get_x and get_all
        methods. Available types of initial points (parameter inittype) are:
            sobol
            random
            grid
        """
        dim = bounds.shape[0]
        self.init_data = None
        if inittype.lower() == "sobol":
            self.init_data = self._sobol(dim, bounds, initpts)
        elif inittype.lower() == "random":
            self.init_data = self._random(dim, bounds, initpts)
        elif inittype.lower() == "grid":
            npts = int(np.power(initpts, 1.0 / dim))
            if initpts != npts ** dim:
                initpts = npts ** dim
                warnings.warn(
                    "Grid based initial point creation modifies"
                    + " initpts so that nth root of it is an integer"
                    + " , where n in the number of dimensions."
                )
            self.init_data = self._grid(bounds, npts)
        else:
            raise TypeError(
                "Unknown option set to keyword inittype. "
                + "Unable to determine initial data."
            )

    def get_x(self, i):
        """
        Returns the i:th initial point
        """
        return self.init_data[i, :]

    def get_all(self):
        """
        Returns all generated initial points
        """
        return self.init_data

    def _sobol(self, dim, bounds, npts):
        """
        Initial points with the quasi-random Sobol sequence
        """
        if npts < 1:
            return  ### STOP here if npts < 1
        sobs = np.transpose(sobol_seq.i4_sobol_generate(dim, npts, 1))
        pts = np.array([]).reshape(0, dim)
        for p in range(npts):
            point = np.array([])
            for d in range(dim):
                a = sobs[p, d] * (bounds[d][1] - bounds[d][0]) + bounds[d][0]
                point = np.append(point, a)
            pts = np.append(pts, [point], axis=0)
        return pts

    def _random(self, dim, bounds, npts):
        """
        Initial points randomly
        """
        pts = []
        for p in range(npts):
            point = []
            for d in range(dim):
                point.append(
                    np.random.rand() * (bounds[d][1] - bounds[d][0]) + bounds[d][0]
                )
            pts.append(point)
        return np.array(pts).astype(float)

    def _grid(self, bounds, npts):
        """
        Initial points in a grid. Total number of points returned is npts^dim.
        """
        base = [np.linspace(*b, num=npts, endpoint=False) for b in bounds]
        return np.array(list(itertools.product(*base))).astype(float)



import numpy as np


def lcb(x, model, params):
    """
    GP-Lower Confidence Bound acquisition function with increasing exploration

    Takes one parameter: the exploration weight.
    """

    explr_weight = params[0]
    m, s, dmdx, dsdx = model.predict_mean_sd_grads(x)
    f_acqu = m - explr_weight * s
    df_acqu = dmdx - explr_weight * dsdx
    scipygradient = np.asmatrix(df_acqu).transpose()
    return f_acqu, scipygradient



import numpy as np


def elcb(x, model, params):
    """GP-Lower Confidence Bound acquisition function with increasing exploration.

    Doesn't take any parameters (apart from the model).
    The exploration weight is given by

    explr_weight = sqrt( 2*log[ ( i^((dim/2) + 2)*pi^(2) ) / ( 3*0.1 ) ] )

    The implementation is based on the following papers

    (1) N. Srinivas, A. Krause, S. M. Kakade, and M. Seeger.  Gaussian process optimization
    in the bandit setting: No regret and experimental design. Proc. ICML, 2010

    (2) E. Brochu, V. M. Cora, and N. de Freitas. A tutorial on Bayesian optimization of expensive
    cost functions, with application to active user modeling and hierarchical reinforcement learning.
    arXiv:1012.2599, 2010.

    where the delta parameter introduced in Brochu et al. has been set to 0.1.
    """
    ndata = model.X.shape[0]
    dim = model.X.shape[1]
    upstairs = (ndata ** ((dim / 2.0) + 2.0)) * (np.pi ** 2.0)
    downstairs = 3 * 0.1
    explr_weight = np.sqrt(2 * np.log10(upstairs / downstairs))

    m, s, dmdx, dsdx = model.predict_mean_sd_grads(x)
    f_acqu = m - explr_weight * s
    df_acqu = dmdx - explr_weight * dsdx
    scipygradient = np.asmatrix(df_acqu).transpose()
    return f_acqu, scipygradient



__version__ = "1.5"
__url__ = "https://gitlab.com/cest-group/boss"
__description__ = "Bayesian optimization structure search"
__authors__ = [
    "Ville Parkkinen",
    "Henri Paulamaki",
    "Arttu Tolvanen",
    "Ulpu Remes",
    "Nuutti Sten",
    "Joakim Loefgren",
    "Milica Todorovic",
]
__maintainer__ = "The BOSS developers team"
__maintainer_email__ = "milica.todorovic@aalto.fi"
__license__ = "Apache License 2.0"



from abc import ABC, abstractmethod
import GPy
import numpy as np


class BaseModel(ABC):
    """
    Base class for surrogate models used in Bayesian optimization.
    """

    # predictions:

    @abstractmethod
    def predict(self, x, noise=True):
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise).
        """
        pass

    @abstractmethod
    def predict_grads(self, x):
        """
        Returns prediction mean and variance gradients with respect to input
        at point x.
        """
        pass

    @abstractmethod
    def predict_mean_sd_grads(self, x, noise=True):
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise).
        """
        pass

    @abstractmethod
    def predict_mean_grad(self, x):
        """
        Returns model mean and its gradient at point x.
        """
        pass

    @abstractmethod
    def estimate_num_local_minima(self, search_bounds):
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        pass

    # model parameters:

    @abstractmethod
    def get_all_params(self):
        """
        Returns model parameters as a dictionary.
        """
        pass

    @abstractmethod
    def get_unfixed_params(self):
        """
        Returns the unfixed parameters of the model in an array.
        """
        pass

    @abstractmethod
    def sample_unfixed_params(self, num_samples):
        """
        Sample unfixed model parameters.
        """
        pass

    @abstractmethod
    def set_unfixed_params(self, params):
        """
        Sets the unfixed parameters of the model to given values.
        """
        pass

    @abstractmethod
    def optimize(self):
        """
        Updates unfixed model parameters.
        """
        pass

    # observations:

    @abstractmethod
    def add_data(self, X_new, Y_new):
        """
        Updates the model evidence (observations) dataset appending.
        """
        pass

    @abstractmethod
    def redefine_data(self, X, Y):
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        pass

    @abstractmethod
    def get_best_xy(self):
        """
        Returns the lowest energy acquisition (x, y).
        """
        pass

    @property
    @abstractmethod
    def kernel(self):
        pass

    @property
    @abstractmethod
    def X(self):
        pass

    @property
    @abstractmethod
    def Y(self):
        pass


class Model(BaseModel):
    """
    Functionality for creating, refitting and optimizing a GP model
    """

    def __init__(self, X, Y, kernel, noise, ynorm):
        """
        Initializes the Model class.
        """
        self.dim = kernel.input_dim
        # normalise observation mean:
        self.normmean = np.mean(Y)
        # scale normalisation is not used unless ynorm is true:
        self.use_norm = ynorm
        # previous boss code used normsd to normalise observation variance:
        # if self.ynorm: self.normsd = np.std(Y)
        # current version normalises observation range:
        self.normsd = np.ptp(Y) if self.use_norm else 1
        # note that the choice betweeen variance or range normalisation needs
        # to be taken into account when we set kernel parameter priors
        # normalised data:
        Y_norm = (Y - self.normmean) / self.normsd
        # initialise model
        self.model = GPy.models.GPRegression(
            X,
            Y_norm,
            kernel=kernel,
            noise_var=noise
        )
        self.model.likelihood.fix()

    # predictions:

    def predict(self, x, noise=True):
        """
        Returns model prediction mean and variance at point x, with or without
        model variance (noise).
        """
        m, v = self.model.predict(np.atleast_2d(x), include_likelihood=noise)
        v = np.clip(v, 1e-12, np.inf)
        return m*self.normsd+self.normmean, v*(self.normsd**2)

    def predict_grads(self, x):
        """
        Returns model prediction mean and variance gradients with respect to
        input at point x.
        """
        dmdx, dvdx = self.model.predictive_gradients(np.atleast_2d(x))
        return dmdx*self.normsd, dvdx*(self.normsd**2)

    def predict_mean_sd_grads(self, x, noise=True):
        """
        Returns the model prediction mean, standard deviation and their
        gradients at point x, with or without model variance (noise).

        This method is a wrapper used primarily during calculations
        of acquisition functions and their derivatives.
        """
        m, v = self.predict(np.atleast_2d(x), noise=noise)
        dmdx, dvdx = self.predict_grads(np.atleast_2d(x))
        dmdx = dmdx[:, :, 0]
        dsdx = dvdx / (2 * np.sqrt(v))
        return m, np.sqrt(v), dmdx, dsdx

    def predict_mean_grad(self, x):
        """Returns model mean and its gradient at point x.

        This method is a wrapper used primarily when the 
        mean function is minimized in order to obtain a
        global minimum prediction.
        """
        m, v = self.predict(np.atleast_2d(x))
        dmdx, dvdx = self.predict_grads(np.atleast_2d(x))
        scipygradient = np.asmatrix(dmdx).transpose()
        return m, scipygradient

    def estimate_num_local_minima(self, search_bounds):
        """
        Returns estimated number of local minima within bounds, calculated
        based on model properties.
        """
        # For the ith dimension, the number of local minima along a slice
        # is approximately n(i) = boundlength(i)/(2*lengthscale(i)). Note
        # that periodic kernels operate on normalised distances: distance
        # between inputs that are period(i)/2 apart is 1. To get the total
        # number of minima for all of the search space, multiply together
        # n(i) over all i.
        numpts = 1
        ks = self.model.kern.parameters if self.dim > 1 else [self.model.kern]
        for bounds, kern in zip(search_bounds, ks):
            if hasattr(kern, 'period'):
                bound_distance = (bounds[1]-bounds[0])/float(kern.period)
            else:
                bound_distance = (bounds[1]-bounds[0])/2
            numpts *= max(1, bound_distance/float(kern.lengthscale))
        return int(numpts)

    # model parameters:

    def get_all_params(self):
        """
        Returns model parameters as a dictionary with entries::
        noise, variance, lengthscales, periods
        where the last two are 1D lists. There exists a period only for those
        dimensions which are using a periodic kernel.
        """
        noise = float(self.model.likelihood.variance)
        sigma = float(self.model.kern.param_array[0])
        lss = []
        pers = []
        ks = self.model.kern.parameters if self.dim > 1 else [self.model.kern]
        for kern in ks:
            lss.append(float(kern.lengthscale))
            if hasattr(kern, 'period'):
                pers.append(float(kern.period))

        # the variables are returned in a dict:
        params = {}
        params['noise'] = noise
        params['variance'] = sigma
        params['lengthscales'] = lss
        params['periods'] = pers

        return params

    def get_unfixed_params(self):
        """
        Returns the unfixed parameters of the model in an array.
        """
        return np.array(self.model.unfixed_param_array.copy()).astype(float)

    def sample_unfixed_params(self, num_samples):
        """
        Sample unfixed model parameters.
        """
        hmc = GPy.inference.mcmc.HMC(self.model)
        burnin = hmc.sample(int(num_samples * 0.33))
        return hmc.sample(num_samples)

    def set_unfixed_params(self, params):
        """
        Sets the unfixed parameters of the model to given values.
        """
        self.model[self.model._fixes_] = params
        self.model.parameters_changed()

    def optimize(self, restarts=1):
        """
        Updates the model hyperparameters by maximizing marginal likelihood.
        """
        self.model.optimization_runs = []
        if restarts == 1:
            self.model.optimize()
        else:
            self.model.optimize_restarts(
                num_restarts=restarts, verbose=False, messages=False
            )

    # observations:

    def add_data(self, X_new, Y_new):
        """
        Updates the model evidence (observations) dataset appending.
        """
        # construct new unnormalised dataset
        X = np.vstack([self.X, np.atleast_2d(X_new)])
        Y = np.vstack([self.Y, Y_new])
        # update model
        self.redefine_data(X, Y)

    def redefine_data(self, X, Y):
        """
        Updates the model evidence (observations) dataset overwriting.
        """
        # update normalisation
        self.normmean = np.mean(Y)
        if self.use_norm:
            self.normsd = np.ptp(Y)
        # update model
        Y_norm = (Y - self.normmean) / self.normsd
        self.model.set_XY(np.atleast_2d(X), np.atleast_2d(Y_norm))

    def get_best_xy(self):
        """
        Returns the lowest energy acquisition (x, y).
        """
        xbest = np.array(self.X[np.argmin(self.Y)])
        ybest = np.min(self.Y)
        return xbest, ybest

    @property
    def kernel(self):
        return self.model.kern

    @property
    def X(self):
        return self.model.X

    @property
    def Y(self):
        return self.model.Y*self.normsd+self.normmean
